package com.aliasi.cluster;

import java.io.*;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;


import com.aliasi.cluster.AbstractHierarchicalClusterer.PairScore;
import com.aliasi.util.ScoredObject;


import com.memetix.mst.detect.Detect;
import com.memetix.mst.language.Language;
import com.memetix.mst.translate.Translate;

/*import edu.sussex.nlp.jws.JiangAndConrath;
import edu.sussex.nlp.jws.Lin;
import edu.sussex.nlp.jws.Path;
import edu.sussex.nlp.jws.WuAndPalmer;
import fi.cs.StringSimilarity; */
import fi.cs.WordnetSimilarity;

public class textAnalyzer {
	WordnetSimilarity wSimilarity=new WordnetSimilarity();

	/**
	 * @param args
	 * @throws IOException
	 */
	public enum INPUT_TYPE {

	    F,  // function
	    IT, // input type
	    IF, // input file
	    OF, // output file
	    NC, // number of clusters
	    NV, // "noun" or "verb"
	    S,  // similarity type of words
	    W1, // word 1
	    W2, // word 2
	    L1, // language 1
	    L2  // language 2
	}
	
	double INFINITY =Double.POSITIVE_INFINITY;
	
	// usage:
	// java -jar textAnalyzer.jar
	// -f n: function... n= 1:keyword clustering 2:similarity of a word list 3:similarity of two words
	//                      4: translate a word 5: language detection
	// -it n: input type for keyword clustering n= 1:words 2:matrix of similarity of words
	// -nc n: number of clusters
	// -nv n: type of words, noun or verb, n= "n":noun, "v":verb
	// -s n: type of similarity metric, n=1: Jiang, n=2: wu, n=3: Lin, n=4: Path 
	// -w1 word: word1
	// -w2 word: word2
	// -l1 language: translate from this language
	// -l2 language: translate to this language

	// output:
	// 1- keyword clustering: text file including labels of words 
	// 2- similarity of a word list: text file including a matrix of similarity of every two words 
	// 3- similarity of two words: similarity value as [x]
	// 4- translation of one word: one word text
	// 5- detection of language of a word
	// -1: one or two words do not exist
	// -2: input parameters are wrong 
	
	public static void main(String[] args) throws IOException {
		// TODO Auto-generated method stub
		
		// parse input arguments
		List<String> argsList = new ArrayList<String>();  
		List<String> optsList = new ArrayList<String>();
		INPUT_TYPE enumval;
		int func = 0;
		int wordSimilarityMethod = 1;
		double [][]similarityMatrix;
		int inputType = 1;
		int numClusters = 0; // if 0 automatic calculation of number of clusters
		String inputFile = ""; // a text file including words or similarities of words
		String outputFile = ""; // a text file
		String wordType = "n";
		String word1 = "";
		String word2 = ""; 
		String lang1 = "FINNISH";
		String lang2 = "ENGLISH";
		String argsT;
		String optsT;
		
		textAnalyzer mainObj = new textAnalyzer();

	    for ( int i = 0; i < args.length; i++ ) {
	        switch ( args[i].charAt(0) ) {
		        case '-':
		            if (args[i].length() < 2)
		                throw new IllegalArgumentException("Not a valid argument: "+args[i]);
		            else {
		                if (args.length-1 == i)
		                    throw new IllegalArgumentException("Expected arg after: "+args[i]);
		                // -opt
		                optsList.add(args[i].substring(1).toUpperCase());
		                argsList.add(args[i+1]);
		                i++;
		            }
		            break;
		        default:
		            break;
	        }
	    }
	    
	    // build input parameters
	    for (int i = 0; i < optsList.size(); i++) {
	    	optsT = optsList.get(i);
	    	argsT = argsList.get(i);
	    	if ( !optsT.equals("F") && !optsT.equals("IT") && !optsT.equals("IF") && !optsT.equals("OF")
	    			&& !optsT.equals("NC") && !optsT.equals("NV") && !optsT.equals("S")
	    			&& !optsT.equals("W1") && !optsT.equals("W2") && !optsT.equals("L1")
	    			&& !optsT.equals("L2") ) {
	    		continue;
	    	}
	    	enumval = INPUT_TYPE.valueOf(optsT);
	    	switch ( enumval ) {
	    	case F:
	    		func = Integer.parseInt(argsT);
	    		break;
	    	case IT:
	    		inputType = Integer.parseInt(argsT);
	    		break;
	    	case IF:
	    		inputFile = argsT.toLowerCase();
	    		break;
	    	case OF:
	    		outputFile = argsT.toLowerCase();
	    		break;
	    	case NC:
	    		numClusters = Integer.parseInt(argsT);
	    		break;
	    	case NV:
	    		wordType = argsT;
	    		break;
	    	case S:
	    		wordSimilarityMethod = Integer.parseInt(argsT);
	    		break;
	    	case W1:
	    		word1 = argsT.toLowerCase();
	    		break;
	    	case W2:
	    		word2 = argsT.toLowerCase();
	    		break;
	    	case L1:
	    		lang1 = argsT.toUpperCase();
	    		break;
	    	case L2:
	    		lang2 = argsT.toUpperCase();
	    		break;
	    	default:
	    		break;
	    	}
	    }
	    
	    System.out.println("Step 1");
	    System.out.println(func);
	    
	    // functions
	    switch (func) {
	    case	1: // keyword clustering of words
	    	numClusters = mainObj.keywordClustering(inputType, inputFile, outputFile, numClusters, wordSimilarityMethod, wordType);
	    	System.out.println(numClusters);
	    	break;
	    case	2: // similarity of a word list
	    	mainObj.similarityWordList(inputFile, outputFile,  wordSimilarityMethod, wordType, 1, null);
	    	System.out.println("Similarity values written in file.");
//	    	System.out.println(numWords);
	    	break;
	    case	3: // similarity of two words
	    	double similariyValue = mainObj.similarityTwoWords(word1, word2, wordSimilarityMethod, wordType);
	    	if ( similariyValue < 0 ) // == -1.0
	    		System.out.println("One of the words or both do not exist in WordNet or\n it is not a \"noun\" or \"verb\"");
	    	else
	    		System.out.println(similariyValue);
	    	break;
	    case	4: // translation of one word
	    	System.out.println("Lang1");
	    	mainObj.translateWord(word1, lang1, lang2);
	    	System.out.println("Lang2");
	    	break;
	    case	5: // detect language of a word
	    	mainObj.detectWordLanguage(word1);
	    	break;
	    default:
	    	throw new IllegalArgumentException("Invalid arguments! ");

	    }
	}
	
	// all methods gives -1.0 if one word or both do not exist in WordNet
	public Double similarityTwoWords(String word1, String word2, int method, String type) {
		double similarityValue = 0;
		try {
			
			//1.Jiang 2.Wu 3.Lin 4.Path 
			if (method ==1 || method ==2 || method ==3 || method==4) {
			
				similarityValue = wSimilarity.getWordnetSimilarity(word1, word2, type, method);
				if ( similarityValue > 1.0 ) // seems only for same words or synonyms gives a big number for Jiang
					similarityValue = 1.0;

			} else{
				System.out.println("Sorry, no this method!!!");
				similarityValue = -2;
			}
		
		} catch (IOException e) {
			
			e.printStackTrace();
		}
		double similarityDouble = similarityValue;
		return similarityDouble;
	}
	
	// finding similarity values between words
	// outputMode: 1: output to file   2: output as matrix (variable) 
	// input: a text file inputfile
	// output: a text file outputFile, just similarity values as matrix
	public double [][] similarityWordList(String inputFile, String outputFile, int method, String wordType, int outputMode, double [][]similarityMatrix){
		int numWords = 0;
		
		textAnalyzer mainObj = new textAnalyzer();
		BufferedReader reader = null;
		BufferedWriter writer = null;
		double similariyValue;
		String word1 = "";
		String word2 = "";
		String []words_temp = new String[1000];
		String []words = null;
		int i = 0;
		int j = 0;
		int nInvalidWords = 0;
		
		try {
			reader = new BufferedReader(new FileReader(inputFile));
			String line = null;
			i = 0;
			while ((line = reader.readLine()) != null) {
				words_temp[i] = line;
				i++;
				if ( i >= 1000 ){
					throw new IllegalArgumentException("Number of words must be less than 1000!");
				}
			}
			
			reader.close();
			
			if ( i == 0 )
				throw new IllegalArgumentException("No words!");
			
			numWords = i;
			
			words = new String[numWords];
			for ( i= 0 ; i < numWords ; i++)
				words[i] = words_temp[i];
			
			if ( outputMode == 1 )
				writer = new BufferedWriter(new FileWriter(outputFile));
			
			if ( outputMode == 2 ) {
				similarityMatrix = new double[numWords][];
				for ( i= 0 ; i < numWords ; i++)
					similarityMatrix[i] = new double[numWords];
			}
			
			for ( i= 0 ; i < numWords ; i++) {
				word1 = words[i];
				for ( j= 0 ; j < numWords ; j++) {
					word2 = words[j];
					similariyValue = mainObj.similarityTwoWords(word1, word2, method, wordType);
					
					if ( outputMode == 1 ) {
						writer.write(String.format("%.3f",similariyValue));
						if ( j < numWords -1 )
							writer.write("\t");
					}
					
					if ( outputMode == 2 )
						similarityMatrix[i][j] = similariyValue;
					
					if ( similariyValue < 0 )
						nInvalidWords++;
				}
				
				if ( (outputMode == 1) && (i < numWords -1) )
					writer.write("\n");
			}
	        
			if ( outputMode == 1 )
				writer.close();
			
			if ( nInvalidWords > 0 )
				System.out.println("Some of the words do not exist in WordNet or\n they are not noun or verb!\n");
	
		} catch (Exception e) {
			
		}
		
		return similarityMatrix;
	}
	
	public String translateWord(String word, String lang1, String lang2) {
		// marketplace: rezaei@cs.uef.fi pass: hvdh....
		// Client ID:     textAnalyzerM100 
		// Client secret: U0CA2Bq8ETZQ1KPEO7Lve4DfkZZyVJtqa3VWGRuOKSk=
		Translate.setClientId("textAnalyzerM100");
		Translate.setClientSecret("U0CA2Bq8ETZQ1KPEO7Lve4DfkZZyVJtqa3VWGRuOKSk=");
		 
		String outString = "x"; 
		try {
			if ( lang1.equals("FINNISH") && lang2.equals("ENGLISH") ) {
				outString = Translate.execute(word, Language.FINNISH, Language.ENGLISH);
				System.out.println(outString);
			}
			else
				System.out.println("Error: Translation only from Finnish to English is supported now!");
		} catch (Exception e) {
			System.out.println("Something wrong!" + e);
		}
		
		return outString;
	}
	
	public String detectWordLanguage(String word) {
		// marketplace: rezaei@cs.uef.fi pass: hvdh....
		// Client ID:     textAnalyzerM100 
		// Client secret: U0CA2Bq8ETZQ1KPEO7Lve4DfkZZyVJtqa3VWGRuOKSk=
		
		Detect.setClientId("textAnalyzerM100");
		Detect.setClientSecret("U0CA2Bq8ETZQ1KPEO7Lve4DfkZZyVJtqa3VWGRuOKSk=");
		
		String outString = "ENGLISH"; 
		String t1;
		String t2;
		String t;
		try {
    		Language detectedLanguage = Detect.execute(word);
    		t = detectedLanguage.toString();
    		if ( t.equals("en")) { // check meaningless word
    			t1 = Translate.execute(word, Language.ENGLISH, Language.FINNISH);
    			t2 = Translate.execute(word, Language.ENGLISH, Language.PERSIAN);
    			t1 = t1.toLowerCase();
    			t1 = t2.toLowerCase();
    			if ( t1.equals(word) && t2.equals(word)) {
    				// first we guess the input word is Finnish
    				// if we specify the input language, bing can translate it for some words
    				t1 = Translate.execute(word, Language.FINNISH, Language.ENGLISH);
    				t1 = t1.toLowerCase();
    				if ( t1.equals(word) )
    					t = "xx";
    				else
    					t = "fi";
    			}
    		}
    		System.out.println(t);
    	} catch (Exception e) {
			System.out.println("Something wrong!" + e);
		}
		
		return outString;
	}
	
	public double [][] readSimilarityMatrixFile(String inputFile, double [][]similarityMatrix){
		int i = 0;
		int j = 0;
		int n = 0;
		int numWords = 0;
		double [][]matrixTemp = null;
		String []temp = null;
		String regexp = "[\\s;\\n\\t]+";
		
		matrixTemp = new double[1000][];
		for ( i= 0 ; i < 1000 ; i++)
			matrixTemp[i] = new double[1000];
		
		BufferedReader reader = null;
		
		try {
			reader = new BufferedReader(new FileReader(inputFile));
			String line = null;
			i = 0;
			while ((line = reader.readLine()) != null) {
				line = line.trim();
				line = line.replaceAll(",", ".");
				temp = line.split(regexp); //split spaces
				
				if ( i==0 )
					n = temp.length;
				else if ( temp.length != n )
					throw new IllegalArgumentException("Invalid simialrity file!");
				
				if ( temp.length >= 1000 || temp.length == 0 )
					throw new IllegalArgumentException("Similarity matrix dimension must be less than 1000*1000!");
				
				for ( j= 0 ; j < temp.length ; j++)
					matrixTemp[i][j] = Double.parseDouble(temp[j]); 
				
				i++;
				if ( i >= 1000 ){
					throw new IllegalArgumentException("Similarity matrix dimension must be less than 1000*1000!");
				}
			}
			
			reader.close();
			
			if ( i == 0 )
				throw new IllegalArgumentException("No words!");
			
			numWords = i;
			
			similarityMatrix = new double[numWords][];
			for ( i= 0 ; i < numWords ; i++)
				similarityMatrix[i] = new double[numWords];
			
			for ( i= 0 ; i < numWords ; i++)
				for ( j= 0 ; j < numWords ; j++)
					similarityMatrix[i][j] = matrixTemp[i][j];
	
		} catch (Exception e) {
			System.out.println("Invalid similarity file!" + e);
		}
		
		
		return similarityMatrix;
	}
	
	public int keywordClustering(int inputType, String inputFile, String outputFile, int numClusters, int wordSimilarityMethod, String wordType){
		int nClusters = 0;
		int numWords = 0;
		int clusteringMethod = 1; // 1: hierarchical clustering 2: k-medoid
		int mergeMethod = 1; // 1: single-link 2: complete-link 3: average-link  
		double [][]similarityMatrix = null;
		int nInvalidWords = 0;
		int i, j;
		
		// calculate similarity matrix of words or directly read it from a file
		if ( inputType == 1 ) { // list of words
			similarityMatrix = similarityWordList(inputFile, "", wordSimilarityMethod, wordType, 2, similarityMatrix);
			numWords = similarityMatrix.length;
		} else if ( inputType == 2 ) { // similarity matrix 
			similarityMatrix = readSimilarityMatrixFile(inputFile, similarityMatrix);
			numWords = similarityMatrix.length;
		} else
			throw new IllegalArgumentException("Invalid input type!");
		
		if ( numWords > 0 ) {
			// check similarity matrix, all the values must be in the range [0, 1]
			for ( i = 0 ; i < numWords ; i++ )
				for ( j = 0 ; j < numWords ; j++ )
					if ( similarityMatrix[i][j] < 0 || similarityMatrix[i][j] > 1 )
						nInvalidWords++;
			
			if ( nInvalidWords > 0 ) {
				System.out.println("Invalid simialrity values!\nAll the values must be in the range of [0, 1].\n");
			} else {
					
				// do keyword clustering using similarity matrix of words
				// methods
				if ( clusteringMethod == 1 ) {
					nClusters = hierarchicalKeywordClustering(similarityMatrix, numClusters, mergeMethod, outputFile);
				}
			}
		}
		
		return nClusters;
	}
	
	public int hierarchicalKeywordClustering(double [][]similarityMatrix, int numClusters, int mergeMethod, String outputFile){
		int numWords = 0;
		int nClusters = 0;
		int i = 0;
		BufferedWriter writer = null;
		int []labels = null;
		numWords = similarityMatrix.length;
		labels = new int[numWords];
		
		if ( numWords > 0 ) {
			labels = new int[numWords];
			
			nClusters = linkClustering(similarityMatrix, mergeMethod, numClusters, labels); // single, complete and average-link
			
		}
		
		try {
			// write data cluster labels to output file
			writer = new BufferedWriter(new FileWriter(outputFile));
			if ( nClusters > 0 ) {
				for ( i = 0 ; i < numWords ; i++ ) {
					writer.write(String.format("%d", labels[i]+1));
					if ( i < numWords -1 )
						writer.write("\n");
				}
			} else
				writer.write("No resulted cluster!");
			
			writer.close();
		} catch (Exception e) {
			System.out.println("Cannot write to file!");
		}
	
		
		return nClusters;
	}
	
	// mergeMethod= 1: single-link   2: complete-link   3: average-link   
	public int linkClustering(double [][]similarityMatrix, int mergeMethod, int numClusters, int [] labels) {
		int N = similarityMatrix.length; // number of objects
		int nClusters = 0;
		int nClustersAuto = 0;
		int	detectNC = 0; // auto detection of number of clusters
		int i, j, i1, i2, c;
		double [][] simTemp = null;
		int [] iMax = null; // index of maximum similar cluster
		double sc = 100;
		double scMin = 100;
		int [] labelsTemp = new int[N];
		
		// number of clusters
		if ( numClusters <= 0 ) { // auto detection of number of clusters
			nClusters = 1;
			detectNC = 1;
		}
		else
			nClusters = numClusters;
		
		if ( numClusters == 1 ) { // no need for clustering, all objects in one cluster
			for ( i = 0 ; i < N ; i++ )
				labels[i] = 0;
		}
		else {
		
			// temporary similarity matrix because we change it
			simTemp = new double[N][];
			for ( i= 0 ; i < N ; i++)
				simTemp[i] = new double[N];
			
			iMax = new int[N];
			
			for ( i = 0 ; i < N ; i++ )
				iMax[i] = i;
			
			for ( i = 0 ; i < N ; i++ )
				for( j = 0 ; j < N ; j++ ) {
					if ( i == j )
						simTemp[i][j] = -1;
					else
						simTemp[i][j] = similarityMatrix[i][j];
					
					if ( simTemp[i][j] > simTemp[i][iMax[i]] )
						iMax[i] = j;
				}
			
			// initialize clusters
			for ( i = 0 ; i < N ; i++ ) {
				labels[i] = i; // each data object in its own cluster
				labelsTemp[i] = i;
			}
			
			nClustersAuto = N;
			scMin = criteriaSC(similarityMatrix, labels, N);
			
			// main loop
			c = N; // number of initial clusters
			while ( c > nClusters ) {
				
				// find two most similar clusters
				i1 =0;
				for( i =0 ; i < N ; i++ )
					if ( simTemp[i][iMax[i]] > simTemp[i1][iMax[i1]] ) 
						i1 = i;
				i2 = iMax[i1]; // we know that i1 is always less than i2
				
				
				// update similarities between all clusters and the new cluster i1
				updateSimilaritiesAfterMerge(simTemp, i1, i2, mergeMethod, labels);
				
				// update cluster labels
				for ( i = 0 ; i < N ; i++ )
					if ( labels[i] == i2 )
						labels[i] = i1;
				
				// (-1)-out old row i2 and column i2
				for ( i = 0 ; i < N ; i++ )
					simTemp[i2][i] = simTemp[i][i2] = -1;
				
				// update iMax and replace ones that previous pointed to i2 to point to i1
				for ( j = 0 ; j < N ; j++ )	{
					if ( iMax[j] == i2 ) 
						iMax[j] = i1;
					if ( simTemp[i1][j] > simTemp[i1][iMax[i1]])
						iMax[i1] = j;
				}
				
				c = c - 1;
				
				if ( detectNC == 1 ) { // auto detection of number of clusters
					// calculate SC criteria
					sc = criteriaSC(similarityMatrix, labels, c);
					if ( sc < scMin ) {
						scMin = sc;
						nClustersAuto = c;
						for ( i = 0 ; i < N ; i++ )
							labelsTemp[i] = labels[i];
					}
				}
				
			}
			
			if ( detectNC == 1 ) { // auto detection of number of clusters
				for ( i = 0 ; i < N ; i++ )
					labels[i] = labelsTemp[i];
				nClusters = nClustersAuto;
			}
		}
		
		return nClusters;
	}
	
	// update similarity between all clusters with new cluster i1 after merging clusters i1 and i2
	// mergeMethod= 1: single-link   2: complete-link   3: average-link 
	public void updateSimilaritiesAfterMerge(double [][]similarityMatrix, int i1, int i2, int mergeMethod, int [] labels) {
		int i, j, N, n1, n2;
		double s;
		
		N = similarityMatrix.length;
		
		for ( i = 0 ; i < N ; i++ ) {
			
			switch ( mergeMethod ) {
			case 1: // single-link
				// overwrite row i1 with maximum of entries in row i1 and i2
				if ( similarityMatrix[i2][i] > similarityMatrix[i1][i] )
					similarityMatrix[i1][i] = similarityMatrix[i][i1] = similarityMatrix[i2][i];
				break;
			case 2: // complete-link
				if ( similarityMatrix[i2][i] < similarityMatrix[i1][i] )
					similarityMatrix[i1][i] = similarityMatrix[i][i1] = similarityMatrix[i2][i];
				break;
			case 3: // average-link
				n1 = 0;
				n2 = 0;
				for ( j = 0 ; j < N ; j++ ) {
					if ( labels[j] == i1 )
						n1++;
					
					if ( labels[j] == i2 )
						n2++;
				}
								
				s = (n1*similarityMatrix[i1][i]+n2*similarityMatrix[i2][i])/(n1+n2);
				similarityMatrix[i1][i] = similarityMatrix[i][i1] = s;
				break;
				
				default:
					break;
			}
		}
		similarityMatrix[i1][i1] = -1;
	}
	
	public double criteriaSC(double [][]similarityMatrix, int [] labels, int numClusters) {
		double 	sc = 100;
		double 	s = 0; // separation
		double 	sMax, sMin, I1;
		double 	c = 100; // compactness
		int 	i, j, k1, k2, N, n1, n2;
		N = similarityMatrix.length;
		int	[] iTemp1 = new int [N];
		int	[] iTemp2 = new int [N];
		
		// separation
		for ( k1 = 0 ; k1 < numClusters ; k1++ ) // loop over clusters
			for ( k2 = 0 ; k2 < numClusters ; k2++ ) { // loop over clusters
				// maximum similarity between clusters k1 and k2
				n1 = 0;
				n2 = 0;
				for ( i = 0 ; i < N ; i++ ) {
					if ( labels[i] == k1 ) {
						iTemp1[n1] = labels[i];
						n1++;
					}
					
					if ( labels[i] == k2 ) {
						iTemp2[n2] = labels[i];
						n2++;
					}
				}
				
				sMax = -1;
				for ( i = 0 ; i < n1 ; i++ ) {
					for ( j = 0 ; j < n2 ; j++ )
						if ( similarityMatrix[iTemp1[i]][iTemp2[j]] > sMax )
							sMax = similarityMatrix[iTemp1[i]][iTemp2[j]];
				}
				s += sMax;
			}
			
		s /= numClusters*(numClusters-1); 
		s = 1-s; // convert to distance
		
		// compactness
		I1 = 0;
		for ( k1 = 0 ; k1 < numClusters ; k1++ ) { // loop over clusters
			// minimum similarity between the objects in cluster k1 
			n1 = 0;
			for ( i = 0 ; i < N ; i++ ) {
				if ( labels[i] == k1 ) {
					iTemp1[n1] = labels[i];
					n1++;
				}
			}
			
			if ( n1 == 1 )
				I1 += 1/N;
			else {
				sMin = 100;
				for ( i = 0 ; i < n1 -1 ; i++ ) {
					for ( j = i+1 ; j < n1 ; j++ )
						if ( similarityMatrix[iTemp1[i]][iTemp1[j]] < sMin )
							sMin = similarityMatrix[iTemp1[i]][iTemp1[j]];
				}
				c = sMin;
			}
		}
		
		if ( c < 2 )
			c = 1-c + I1; // convert to distance
		else
			c = I1;
		
		// sc criteria
		sc = c/s;
		
		return sc;
	}
	
}


